/*
 * DiskIO.c - Low-level functions allowing direct disk access under both
 *	Windows 9x and Windows NT.
 *
 *	SectorRead
 *	SectorWrite
 *
 * Copyright 2002 William R. Swanson
 * site@swansontec.com
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *	If you disagree with the terms of the GPL, you may contact me and
 *	request a non-GPL copy of this code for integration into independent
 *	programs not connected with Tweak Disk.
 */

#include "Main.h"

/******************************************************************************
 * Constants and Data Types
 */

#define VWIN32_DIOC_DOS_INT25 (2)
#define VWIN32_DIOC_DOS_INT26 (3)
#define REG_FLAGS_CARRY	(0x0001)

typedef struct _DIOC_REGISTERS {
    DWORD EBX;
    DWORD EDX;
    DWORD ECX;
    DWORD EAX;
    DWORD EDI;
    DWORD ESI;
    DWORD flags;
} DIOC_REGISTERS, *PDIOC_REGISTERS;

/******************************************************************************
 * Private Functions
 */

static long SectorRead9x (char buffer[], long sector, long count, int disk, HANDLE vwin32);
static long SectorReadNT (char buffer[], long sector, long count, int disk, HANDLE device);

static long SectorWrite9x (char buffer[], long sector, long count, int disk, HANDLE vwin32);
static long SectorWriteNT (char buffer[], long sector, long count, int disk, HANDLE device);

/******************************************************************************
 * SectorRead - Reads sectors from a logical drive.
 *
 * Preconditions:
 *	buffer - Buffer to hold read data.
 *	sector - Logical number of first sector to read (<65536).
 *	count - Number of sectors to read (<65536).
 *	disk - Number of logical drive.
 * Postconditons:
 *	buffer - Data read from disk.
 *	return - Number of sectors actually read, != count if error.
 */

long SectorRead (char buffer[], long sector, long count, int disk)
{
	HANDLE device;
	HANDLE vwin32;
	char name[] = "\\\\.\\A:";
	
	vwin32 = CreateFile("\\\\.\\vwin32",
		0,
		0,
		NULL,
		0,
		FILE_FLAG_DELETE_ON_CLOSE,
		NULL);
	if(vwin32 != INVALID_HANDLE_VALUE)
		return SectorRead9x (buffer, sector, count, disk, vwin32);
	
	name[4] += disk;
	device = CreateFile(name,
		GENERIC_READ,
		FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL);
	if(device != INVALID_HANDLE_VALUE)
		return SectorReadNT (buffer, sector, count, disk, device);

	ErrorAPI(APP_NAME);
	return 0;	
}

/******************************************************************************
 * SectorRead9x - Reads a sector from the disk using Windows 9x mechanisms.
 */

static long SectorRead9x (char buffer[], long sector, long count, int disk, HANDLE vwin32)
{
	DWORD cb;
	DIOC_REGISTERS reg;

	reg.EAX = (DWORD)((BYTE)disk);
	reg.EBX = (DWORD)buffer;
	reg.ECX = (DWORD)count;
	reg.EDX = (DWORD)sector;
	if(!DeviceIoControl(vwin32,
		VWIN32_DIOC_DOS_INT25,
		&reg, sizeof(reg), 
		&reg, sizeof(reg), &cb,
		NULL)){
		ErrorAPI(APP_NAME);
		return 0;
	}
	CloseHandle(vwin32);
	if(reg.flags & REG_FLAGS_CARRY){
	    ErrorBox(APP_NAME, ERR_TEXT_DISK);
	    return 0;
	}
	return count;	
}

/******************************************************************************
 * SectorReadNT - Reads a sector from the disk using Windows NT mechanisms.
 */

static long SectorReadNT (char buffer[], long sector, long count, int disk, HANDLE device)
{
	char root[] = "A:\\";
	DWORD sector_size;
	DWORD position;
	DWORD nread;
	BOOL success;

	root[0] += disk;
	if(!GetDiskFreeSpace(root, &nread, &sector_size, &nread, &nread)){
		ErrorAPI(APP_NAME);
		return 0;
	}
	position = SetFilePointer(device, sector * sector_size, NULL, FILE_BEGIN);
	if(position != sector * sector_size){
		ErrorAPI(APP_NAME);
		return 0;
	}
	success = ReadFile (device,
		buffer,
		count * sector_size,
		&nread,
		NULL
		);
	if(!success || nread != count * sector_size){
		ErrorAPI(APP_NAME);
	}
	CloseHandle(device);
	return nread / sector_size;
}

/******************************************************************************
 * SectorWrite - Writes sectors to a physical disk.
 *
 * Preconditions:
 *	buffer - Buffer holding data to write.
 *	sector - Logical number of first sector to write (<65536).
 *	count - Number of sectors to write (<65536).
 *	disk - Number of logical drive.
 * Postconditons:
 *	return - Number of sectors actually written, != count if error.
 */

long SectorWrite (char buffer[], long sector, long count, int disk)
{
	HANDLE device;
	HANDLE vwin32;
	char name[] = "\\\\.\\A:";
	
	vwin32 = CreateFile("\\\\.\\vwin32",
		0,
		0,
		NULL,
		0,
		FILE_FLAG_DELETE_ON_CLOSE,
		NULL);
	if(vwin32 != INVALID_HANDLE_VALUE)
		return SectorWrite9x (buffer, sector, count, disk, vwin32);
	
	name[4] += disk;
	device = CreateFile(name,
		GENERIC_WRITE,
		FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL);
	if(device != INVALID_HANDLE_VALUE)
		return SectorWriteNT (buffer, sector, count, disk, device);

	ErrorAPI(APP_NAME);
	return 0;	
}

/******************************************************************************
 * SectorWrite9x - Writes a sector to the disk using Windows 9x mechanisms.
 */

static long SectorWrite9x (char buffer[], long sector, long count, int disk, HANDLE vwin32)
{
	DWORD cb;
	DIOC_REGISTERS reg;

	reg.EAX = (DWORD)((BYTE)disk);
	reg.EBX = (DWORD)buffer;
	reg.ECX = (DWORD)count;
	reg.EDX = (DWORD)sector;
	if(!DeviceIoControl(vwin32,
		VWIN32_DIOC_DOS_INT26,
		&reg, sizeof(reg), 
		&reg, sizeof(reg), &cb,
		NULL)){
		ErrorAPI(APP_NAME);
		return 0;
	}
	CloseHandle(vwin32);
	if(reg.flags & REG_FLAGS_CARRY){
	    ErrorBox(APP_NAME, ERR_TEXT_DISK);
	    return 0;
	}
	return count;	
}

/******************************************************************************
 * SectorWriteNT - Writes a sector to the disk using Windows NT mechanisms.
 */

static long SectorWriteNT (char buffer[], long sector, long count, int disk, HANDLE device)
{
	char root[] = "A:\\";
	DWORD sector_size;
	DWORD position;
	DWORD nwritten;
	BOOL success;

	root[0] += disk;
	if(!GetDiskFreeSpace(root, &nwritten, &sector_size, &nwritten, &nwritten)){
		ErrorAPI(APP_NAME);
		return 0;
	}
	position = SetFilePointer(device, sector * sector_size, NULL, FILE_BEGIN);
	if(position != sector * sector_size){
		ErrorAPI(APP_NAME);
		return 0;
	}
	success = WriteFile (device,
		buffer,
		count * sector_size,
		&nwritten,
		NULL
		);
	if(!success || nwritten != count * sector_size){
		ErrorAPI(APP_NAME);
	}
	CloseHandle(device);
	return nwritten / sector_size;
}